#include "stdafx.h"
#include "BITSWrappers.h"
#include <array>


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


#pragma warning(suppress: 26426)
ATL::CComModule _Module;

constexpr const wchar_t* g_pwszJobDisplayName = L"BITS Sample Job";


struct CBITSStateLookup
{
  BG_JOB_STATE state;
  const wchar_t* pszState;
};

constexpr std::array<CBITSStateLookup, 9> g_StateLookup
{ {
  { BG_JOB_STATE_QUEUED,          L"Queued"          },
  { BG_JOB_STATE_CONNECTING,      L"Connecting"      },
  { BG_JOB_STATE_TRANSFERRING,    L"Transferring"    },
  { BG_JOB_STATE_SUSPENDED,       L"Suspended"       },
  { BG_JOB_STATE_ERROR,           L"Error"           },
  { BG_JOB_STATE_TRANSIENT_ERROR, L"Transient Error" },
  { BG_JOB_STATE_TRANSFERRED,     L"Transferred"     },
  { BG_JOB_STATE_ACKNOWLEDGED,    L"Acknowledged"    },
  { BG_JOB_STATE_CANCELLED,       L"Cancelled"       }
} };

const wchar_t* BITSStateToString(_In_ BG_JOB_STATE state)
{
  const auto iter = std::find_if(g_StateLookup.cbegin(), g_StateLookup.cend(), [state](const CBITSStateLookup& element) noexcept
    {
      return element.state == state;
#pragma warning(suppress: 26812)
    });

  if (iter != g_StateLookup.cend())
    return iter->pszState;
  else
    return nullptr;
}

class CMyCallback : public BITS::CCallback //You can also try deriving from BITS::CCallback2 or BITS::CCallback3
{
public:
#pragma warning(suppress: 26455)
  CMyCallback() = default;
  CMyCallback(const CMyCallback&) = delete;
  CMyCallback(CMyCallback&&) = delete;
#pragma warning(suppress: 26477)
  ~CMyCallback() = default; //NOLINT(modernize-use-override)
  CMyCallback& operator=(const CMyCallback&) = delete;
  CMyCallback& operator=(CMyCallback&&) = delete;

  HRESULT OnJobError(_In_ BITS::CJob& job, _In_ BITS::CError& error) override
  {
    BITS::String sDisplayName;
    HRESULT hr{job.GetDisplayName(sDisplayName)}; //NOLINT(clang-analyzer-deadcode.DeadStores, clang-diagnostic-unused-but-set-variable)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    BITS::String sErrorDescription;
    hr = error.GetErrorDescription(sErrorDescription); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    BITS::String sContextDescription;
    hr = error.GetErrorContextDescription(sContextDescription); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    BITS::String sProtocol;
    hr = error.GetProtocol(sProtocol); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));

    wprintf(L"OnJobError called, Description:%s, Error Description:%s, Error Context:%s, Protocol:%s\n", sDisplayName.c_str(), sErrorDescription.c_str(), sContextDescription.c_str(), sProtocol.c_str());
    return S_OK;
  }

  HRESULT OnJobModification(_In_ BITS::CJob& job, _In_ DWORD /*dwReserved*/) override
  {
    BITS::String sDisplayName;
    const HRESULT hr{job.GetDisplayName(sDisplayName)}; //NOLINT(clang-analyzer-deadcode.DeadStores, clang-diagnostic-unused-variable)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    wprintf(L"OnJobModification called, Job Description:%s\n", sDisplayName.c_str());
    return S_OK;
  }

  HRESULT OnJobTransferred(_In_ BITS::CJob& job) override
  {
    BITS::String sDisplayName;
    const HRESULT hr{job.GetDisplayName(sDisplayName)}; //NOLINT(clang-analyzer-deadcode.DeadStores, clang-diagnostic-unused-variable)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    wprintf(L"OnTransferred called, Job Description:%s\n", sDisplayName.c_str());
    return S_OK;
  }

  /*
  HRESULT OnFileTransferred(_In_ BITS::CJob& job, _In_ BITS::CFile& file) override
  {
    UNREFERENCED_PARAMETER(file);
    BITS::String sDisplayName;
    HRESULT hr{job.GetDisplayName(sDisplayName)};
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    wprintf(L"OnFileTransferred called, Job Description:%s\n", sDisplayName.c_str());
    return S_OK;
  }

  HRESULT OnFileRangesTransferred(_In_ BITS::CJob& job, _In_ BITS::CFile& file, _In_ DWORD rangeCount, __RPC__in_ecount_full(rangeCount) const BG_FILE_RANGE ranges[]) override
  {
    UNREFERENCED_PARAMETER(file);
    UNREFERENCED_PARAMETER(rangeCount);
    UNREFERENCED_PARAMETER(ranges);
    BITS::String sDisplayName;
    HRESULT hr{job.GetDisplayName(sDisplayName)};
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    wprintf(L"OnFileRangesTransferred called, Job Description:%s\n", sDisplayName.c_str());
    return S_OK;
  }
  */
};

__if_exists(BITS::CServerCertificateValidationCallback)
{
class CMyServerCertificateValidationCallback : public BITS::CServerCertificateValidationCallback
{
public:
  CMyServerCertificateValidationCallback() = default;
  CMyServerCertificateValidationCallback(const CMyServerCertificateValidationCallback&) = delete;
  CMyServerCertificateValidationCallback(CMyServerCertificateValidationCallback&&) = delete;
  ~CMyServerCertificateValidationCallback() = default;
  CMyServerCertificateValidationCallback& operator=(const CMyServerCertificateValidationCallback&) = delete;
  CMyServerCertificateValidationCallback& operator=(CMyServerCertificateValidationCallback&&) = delete;

#pragma warning(suppress: 26485)
  HRESULT OnValidateServerCertificate(_In_ BITS::CJob& job, _In_ BITS::CFile& file, DWORD certLength, const BYTE certData[], DWORD certEncodingType, DWORD certStoreLength, const BYTE certStoreData[]) override
  {
    return __super::OnValidateServerCertificate(job, file, certLength, certData, certEncodingType, certStoreLength, certStoreData);
  }
};
} //__if_exists(BITS::CServerCertificateValidationCallback)

#pragma warning(suppress: 26461 26485)
int DemoCode(int /*argc*/, TCHAR* argv[]) //NOLINT(modernize-avoid-c-arrays)
{
  BITS::CManager mgr;
  _tprintf(_T("Initializing BITS Manager\n"));
  HRESULT hr{mgr.Initialize()}; //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(SUCCEEDED(hr));

  BITS::JobsArray jobs;
  _tprintf(_T("Enumerating BITS jobs\n"));
  hr = mgr.EnumJobs(0, jobs); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSUME(SUCCEEDED(hr));

  //If there are no jobs create one
  const auto nJobs{jobs.size()};
  if (nJobs == 0)
  {
    GUID JobID;
    BITS::CJob job;
    wprintf(L"Creating Job with display name \"%s\"\n", g_pwszJobDisplayName);
    hr = mgr.CreateJob(g_pwszJobDisplayName, BG_JOB_TYPE_DOWNLOAD, JobID, job);
    if (FAILED(hr))
    {
      BITS::String sErrorDescription;
      mgr.GetErrorDescription(hr, sErrorDescription);
      return 1;
    }
    //Add the files to the job
    _tprintf(_T("Adding file to download\n"));
#pragma warning(suppress: 26481 26486)
    hr = job->AddFile(ATL::CT2W{argv[1]}, ATL::CT2W{argv[2]}); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));

    //Lets use foreground priority for this sample app
    hr = job->SetPriority(BG_JOB_PRIORITY_FOREGROUND); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));

    //BITS::CJob3 job3;
    //hr = job.GetJob3(job3); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    //ATLASSUME(SUCCEEDED(hr));
    //BG_FILE_RANGE range;
    //range.InitialOffset = 400;
    //range.Length = 409600;
    //hr = job3->AddFileWithRanges(CT2W{argv[1]}, CT2W{argv[2]}, 1, &range);
#pragma warning(suppress: 26477)
    //ATLASSUME(SUCCEEDED(hr));

    //Setup the callback interface
    _tprintf(_T("Setting notify interface\n"));
    hr = job.SetNotifyInterface<CMyCallback>();
    if (FAILED(hr))
      return 1;
    hr = job->SetNotifyFlags(BG_NOTIFY_JOB_TRANSFERRED | BG_NOTIFY_JOB_ERROR | BG_NOTIFY_JOB_MODIFICATION | BG_NOTIFY_FILE_TRANSFERRED); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));

    //Start the job
    _tprintf(_T("Resuming job\n"));
    hr = job->Resume(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));

    BITS::CJob foundJob;
    hr = mgr.GetJob(JobID, foundJob); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));
    BITS::String sDisplayName;
    hr = foundJob.GetDisplayName(sDisplayName); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
    ATLASSUME(SUCCEEDED(hr));

    //Poll the job for state/progress
    bool bContinue{true};
    while (bContinue)
    {
      Sleep(2500);
      BG_JOB_STATE state;
      hr = job->GetState(&state); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));

      if (state == BG_JOB_STATE_TRANSFERRED)
      {
        _tprintf(_T("Completing job\n"));
        hr = job->Complete(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        bContinue = false;
      }
      else if ((state == BG_JOB_STATE_ERROR) || (state == BG_JOB_STATE_TRANSIENT_ERROR))
      {
        BITS::CError error;
        hr = job->GetError(&error); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
#pragma warning(suppress: 26489)
        BITS::String sErrorDescription;
        hr = error.GetErrorDescription(sErrorDescription); //NOLINT(clang-analyzer-deadcode.DeadStores)

#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        BITS::String sContextDescription;
        hr = error.GetErrorContextDescription(sContextDescription); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        BITS::String sProtocol;
        hr = error.GetProtocol(sProtocol); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        wprintf(L"State:%s, Error Description:%s, Error Context:%s, Protocol:%s\n", BITSStateToString(state), sErrorDescription.c_str(), sContextDescription.c_str(), sProtocol.c_str());

        _tprintf(_T("Attempting to resume job\n"));
        hr = job->Resume(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
      }
      else if (state == BG_JOB_STATE_TRANSFERRING)
      {
        BG_JOB_PROGRESS progress;
        hr = job->GetProgress(&progress); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        if (progress.BytesTotal == BG_SIZE_UNKNOWN)
          wprintf(L"State:%s, File %u of %u, %I64u bytes of Unknown\n", BITSStateToString(state), progress.FilesTransferred + 1, progress.FilesTotal, progress.BytesTransferred);
        else
          wprintf(L"State:%s, File %u of %u, %I64u bytes of %I64u bytes\n", BITSStateToString(state), progress.FilesTransferred + 1, progress.FilesTotal, progress.BytesTransferred, progress.BytesTotal);
      }
      else
          wprintf(L"State: %s\n", BITSStateToString(state));
    }
  }
  else
  {
    wprintf(L"Enumerating Jobs\n");
    for (auto& job : jobs)
    {
      BITS::CJob2 job2;
      hr = job.GetJob2(job2); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      BITS::String sProgram;
      BITS::String sParameters;
      hr = job2.GetNotifyCmdLine(sProgram, sParameters); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));

      BITS::CJob3 job3;
      hr = job.GetJob3(job3); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      BITS::CJob4 job4;
      hr = job.GetJob4(job4); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      hr = S_OK; //NOLINT(clang-analyzer-deadcode.DeadStores)
    __if_exists(BITS::CJob5)
    {
      BITS::CJob5 job5;
      hr = job.GetJob5(job5);
    }
      BITS::String sErrorDescription;
      hr = mgr.GetErrorDescription(hr, sErrorDescription); //NOLINT(clang-analyzer-deadcode.DeadStores)

      BITS::CJobOptions jobOptions;
      hr = job.GetJobOptions(jobOptions); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      BG_CERT_STORE_LOCATION location;
      BITS::String sStoreName;
      BITS::ByteArray certHashBlob;
      BITS::String sSubjectName;
      hr = jobOptions.GetClientCertificate(location, sStoreName, certHashBlob, sSubjectName); //NOLINT(clang-analyzer-deadcode.DeadStores)
      BITS::String sRequestHeaders;
      hr = jobOptions.GetCustomHeaders(sRequestHeaders); //NOLINT(clang-analyzer-deadcode.DeadStores)

    __if_exists(BITS::CJobOptions2)
    {
      BITS::CJobOptions2 jobOptions2;
      hr = job.GetJobOptions2(jobOptions2);
      if (SUCCEEDED(hr))
      {
        BITS::String sMethod;
        hr = jobOptions2.GetHttpMethod(sMethod);
      }
    }

    __if_exists(BITS::CJobOptions3)
    {
      BITS::CJobOptions3 jobOptions3;
      hr = job.GetJobOptions3(jobOptions3);
      if (SUCCEEDED(hr))
      {
        //Setup the callback interface
        hr = jobOptions3.SetServerCertificateValidationInterface<CMyCallback>();
      }
    }

    __if_exists(BITS::CTokenOptions)
    {
      BITS::CTokenOptions tokenOptions;
      hr = job.GetTokenOptions(tokenOptions); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      BITS::String sSid;
      hr = tokenOptions.GetHelperTokenSid(sSid); //NOLINT(clang-analyzer-deadcode.DeadStores)
    }

      BITS::String sDescription;
      hr = job.GetDescription(sDescription); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      wprintf(L"Description: %s\n", sDescription.c_str());

      BITS::String sDisplayName;
      hr = job.GetDisplayName(sDisplayName); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));
      wprintf(L"Display Name: %s\n", sDisplayName.c_str());

      GUID JobID{};
      hr = job->GetId(&JobID); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));

      BITS::FilesArray files;
      hr = job.EnumFiles(files); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));

      BITS::String sOwner;
      hr = job.GetOwner(sOwner); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));

      BG_JOB_PROXY_USAGE proxyUsage;
      BITS::String sProxyList;
      BITS::String sProxyBypassList;
      hr = job.GetProxySettings(proxyUsage, sProxyList, sProxyBypassList); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
      ATLASSUME(SUCCEEDED(hr));

      for (auto& file : files)
      {
        BITS::CFile2 file2;
        hr = file.GetFile2(file2); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));

        BITS::FileRangeArray ranges;
        hr = file2.GetFileRanges(ranges); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));

        BITS::CFile3 file3;
        hr = file.GetFile3(file3); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));

      __if_exists(BITS::CFile4)
      {
        BITS::CFile4 file4;
        hr = file.GetFile4(file4); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
      }

      __if_exists(BITS::CFile5)
      {
          BITS::CFile5 file5;
          hr = file.GetFile5(file5); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
          ATLASSUME(SUCCEEDED(hr));
      }

      __if_exists(BITS::CFile6)
      {
        BITS::CFile6 file6;
        hr = file.GetFile6(file6); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        BITS::FileRangeArray filledRanges;
        hr = file6.GetFilledFileRanges(filledRanges); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
      }

        BITS::String sRemoteName;
        hr = file.GetRemoteName(sRemoteName); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
        BITS::String sLocalName;
        hr = file.GetLocalName(sLocalName); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
      }

      if (sDisplayName == g_pwszJobDisplayName)
      {
        wprintf(L"Cancelling Job as its Display name is \"%s\"\n", g_pwszJobDisplayName);
        hr = job->Cancel(); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
        ATLASSUME(SUCCEEDED(hr));
      }
    }
  }

  return 0;
}

#pragma warning(suppress: 26485 26487)
int _tmain(int argc, TCHAR* argv[]) //NOLINT(modernize-avoid-c-arrays)
{
  if (argc != 3)
  {
    _ftprintf(stderr, _T("BITSWrapper URLToDownload LocalFileToDownloadInto\n"));
    return 1;
  }

  //Initialise COM
  const HRESULT hr{CoInitialize(nullptr)};
  if (FAILED(hr))
  {
#pragma warning(suppress: 26472)
    _ftprintf(stderr, _T("Failed to initialize COM. Error:%08X\n"), static_cast<unsigned int>(hr));
    return 1;
  }

  //Call the function to demo the classes
  DemoCode(argc, argv);

  //Closedown COM
  CoUninitialize();

  return 0;
}
